#######################################################################
#                              Timestep                               #
#######################################################################

function compute_maxspeed(::Project{:valencia1d}, equation, mesh)
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)
end
function compute_maxspeed(::Project{:spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)
end
function compute_maxspeed(::Project{:rescaled_spherical1d}, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
end
function compute_maxspeed(::Project{:doublecartoon}, equation, mesh)
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)
end
function compute_maxspeed(::Project{:cartoon}, equation, mesh)
  broadcast_volume!(maxspeed_cartoon, equation, mesh)
end

function get_maxspeed(mesh, equation)
  @unpack max_v = get_static_variables(mesh.cache)
  vmax = maximum(abs, max_v)
  vmax_limit = 0.99999999999
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed $vmax exceeding $vmax_limit"
    vmax = vmax_limit
  end
  return vmax
end

function timestep(mesh, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K
  dt = dL / (vmax * dtfactor(mesh))
  return dt
end
dtfactor(mesh::Mesh1d{FVElement}) = 2
dtfactor(mesh::Mesh1d{SpectralElement}) = 1

function timestep(mesh::Mesh2d{SpectralElement}, P::Project, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack element = mesh
  @unpack N, z = element
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  Lx, Ly = dg1d.widths(mesh.boxes[1])
  Kx, Ky = mesh.tree.dims
  dL = min(Lx/Kx, Ly/Ky)
  dt = dL / (vmax * dtfactor(mesh))
  return dt
end


function timestep(mesh, P::Project, hrsc::HRSC.AbstractArtificialViscosity)
  @unpack equation = P
  @unpack smoothed_mu = get_static_variables(mesh)
  @unpack element = mesh
  @unpack N = element
  compute_maxspeed(P, P.equation, mesh)
  vmax = get_maxspeed(mesh, P.equation)
  mumax = maximum(abs, smoothed_mu)
  L, = dg1d.widths(mesh)
  K, = mesh.tree.dims
  dL = L/K
  dl = min_grid_spacing(mesh)
  dt1 = dl / (vmax * N)
  dt2 = dL^2 / (mumax * N^3)
  dt = min(dt1, dt2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


#######################################################################
#                               FV rhs                                #
#######################################################################


function rhs!(mesh::Mesh1d{FVElement}, P::Project)
  if P.prms.hrsc === :fv_central
    rhs_fv_central!(mesh, P)
  elseif P.prms.hrsc === :slope_limiter
    rhs_fv_slope_limiter!(mesh, P)
  else
    TODO(P.prms.hrsc)
  end

end

function rhs_fv_central!(mesh, P::Project{:spherical1d})

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                = get_dynamic_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, ρ, ϵ, vr, p,
          src_D, src_Sr, src_τ    = get_static_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ    = get_rhs_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)
  @unpack dt                      = get_global_variables(cache)

  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux, equation, mesh)

  fv_update_step!(rhs_D,  D,  flr_D,  src_D,  bdry_D,  nflr_D,  dt[1], mesh)
  fv_update_step!(rhs_Sr, Sr, flr_Sr, src_Sr, bdry_Sr, nflr_Sr, dt[1], mesh)
  fv_update_step!(rhs_τ,  τ,  flr_τ,  src_τ,  bdry_τ,  nflr_τ,  dt[1], mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_spherical1d, P.equation, mesh)
  end

  return
end


function rhs_fv_central!(mesh, P::Project{:rescaled_spherical1d})

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                = get_dynamic_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          D, Sr, τ,
          src_rD, src_rSr, src_rτ    = get_static_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ    = get_rhs_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_rescaled_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_bdry!(fv_bdry_flux_rescaled_spherical1d, equation, mesh)

  dt = timestep(mesh, P, hrsc)
  fv_update_step!(rhs_rD,  rD,  flr_rD,  src_rD,  bdry_D,  nflr_rD,  dt, mesh)
  fv_update_step!(rhs_rSr, rSr, flr_rSr, src_rSr, bdry_Sr, nflr_rSr, dt, mesh)
  fv_update_step!(rhs_rτ,  rτ,  flr_rτ,  src_rτ,  bdry_τ,  nflr_rτ,  dt, mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_rescaled_spherical1d, P.equation, mesh)
  end

  return
end


function rhs_fv_central!(mesh, P::Project{:doublecartoon})

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rτ                = get_dynamic_variables(cache)
  @unpack flx_rDx, flx_rSxx, flx_rτx,
          D, Sx, τ,
          src_rD, src_rSx, src_rτ    = get_static_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rτ    = get_rhs_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rτ,
          bdry_rD, bdry_rSx, bdry_rτ = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)
  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, equation, mesh)
  broadcast_bdry!(fv_bdry_flux_doublecartoon_tov_mirrored, equation, mesh)

  dt = timestep(mesh, P, P.hrsc)
  if :rD ∉ P.prms.freeze_vars
    fv_update_step!(rhs_rD,  rD,  flx_rDx,  src_rD,  bdry_rD,  nflx_rD,  dt, mesh)
  end
  if :rSx ∉ P.prms.freeze_vars
    fv_update_step!(rhs_rSx, rSx, flx_rSxx, src_rSx, bdry_rSx, nflx_rSx, dt, mesh)
  end
  if :rτ ∉ P.prms.freeze_vars
    fv_update_step!(rhs_rτ,  rτ,  flx_rτx,  src_rτ,  bdry_rτ,  nflx_rτ,  dt, mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_doublecartoon, P.equation, mesh)
  end

  return
end


"""
Helper struct to define buffer values at boundaries and to enable uniform indexing.
"""
mutable struct BdryBufferedData
  bc_ll::Float64
  bc_l::Float64
  data::Vector{Float64}
  bc_r::Float64
  bc_rr::Float64
end

@inline function Base.getindex(d::BdryBufferedData, i)
  N = length(d.data)
  if 1 ≤ i ≤ N
    @inbounds return d.data[i]
  elseif i == 0
    return d.bc_l
  elseif i == -1
    return d.bc_ll
  elseif i == N+1
    return d.bc_r
  elseif i == N+2
    return d.bc_rr
  end
end
@inline function Base.setindex!(d::BdryBufferedData, val::Float64, i)
  N = length(d.data)
  if 1 ≤ i ≤ N
    @inbounds d.data[i] = val
  elseif i == 0
    d.bc_l = val
  elseif i == -1
    d.bc_ll = val
  elseif i == N+1
    d.bc_r = val
  elseif i == N+2
    d.bc_rr = val
  end
end

function rhs_fv_slope_limiter!(mesh, P::Project{:spherical1d})
  @unpack equation             = P
  @unpack D, Sr, τ             = get_dynamic_variables(mesh)
  @unpack D1, Sr1, τ1,
          D2, Sr2, τ2          = get_static_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ = get_rhs_variables(mesh)
  @unpack dt = get_global_variables(mesh)
  dt = dt[1]

  # SSPRK3 stepping from Hesthaven 2018 book, sec 10.2 (cf Script 10.13)
  step_fv_slope_limiter!(mesh, P, (D,Sr,τ))
  if :D ∉ P.prms.freeze_vars
    @. D1  = D  + dt*rhs_D
  end
  if :Sr ∉ P.prms.freeze_vars
    @. Sr1 = Sr + dt*rhs_Sr
  end
  if :τ ∉ P.prms.freeze_vars
    @. τ1  = τ  + dt*rhs_τ
  end

  step_fv_slope_limiter!(mesh, P, (D1,Sr1,τ1))
  if :D ∉ P.prms.freeze_vars
    @. D2  = (3*D  + D1  + dt*rhs_D)/4
  end
  if :Sr ∉ P.prms.freeze_vars
    @. Sr2 = (3*Sr + Sr1 + dt*rhs_Sr)/4
  end
  if :τ ∉ P.prms.freeze_vars
    @. τ2  = (3*τ  + τ1  + dt*rhs_τ)/4
  end

  step_fv_slope_limiter!(mesh, P, (D2,Sr2,τ2))
  # atm final result is published via rhs_ variables
  if :D ∉ P.prms.freeze_vars
    @. rhs_D  = (D  + 2*D2  + 2*dt*rhs_D)/3
  end
  if :Sr ∉ P.prms.freeze_vars
    @. rhs_Sr = (Sr + 2*Sr2 + 2*dt*rhs_Sr)/3
  end
  if :τ ∉ P.prms.freeze_vars
    @. rhs_τ  = (τ  + 2*τ2  + 2*dt*rhs_τ)/3
  end

  return
end


function extrapolate_quadratic(vs, rs, rng, rr1, rr2)
  v1, v2, v3 = vs[rng]
  r1, r2, r3 = rs[rng]
  # utilize the lagrange interpolation formulae for extrapolation
  vv1 = ( (rr1-r2)*(rr1-r3)/((r1-r2)*(r1-r3))*v1 +
          (rr1-r1)*(rr1-r3)/((r2-r1)*(r2-r3))*v2 +
          (rr1-r1)*(rr1-r2)/((r3-r1)*(r3-r2))*v3 )
  vv2 = ( (rr2-r2)*(rr2-r3)/((r1-r2)*(r1-r3))*v1 +
          (rr2-r1)*(rr2-r3)/((r2-r1)*(r2-r3))*v2 +
          (rr2-r1)*(rr2-r2)/((r3-r1)*(r3-r2))*v3 )
  return vv1, vv2
end


function step_fv_slope_limiter!(mesh, P::Project{:spherical1d}, state)

  @unpack equation                      = P
  @unpack D, Sr, τ                      = get_dynamic_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ          = get_rhs_variables(mesh)
  @unpack bD, bSr, bτ,
          ρ, vr, ϵ,
          ρL, vrL, ϵL, pL,
          ρR, vrR, ϵR, pR,
          dρ_p, dvr_p, dϵ_p,
          dρ_m, dvr_m, dϵ_m,
          nflx_D_L, nflx_Sr_L, nflx_τ_L,
          nflx_D_R, nflx_Sr_R, nflx_τ_R,
          src_D, src_Sr, src_τ,
          γrr, α, βru                   = get_static_variables(mesh)
  @unpack eos                           = equation

  # this is a hack
  # the problem is that all @with_signature methods are tied to the
  # dynamic_variables of mesh.cache (D,Sr,τ)
  # but because we don't use the ordinary evolve interface, we have
  # to handle the state vector here manually
  # this requires us to first backup the current dynamic_variables,
  # copy the new state vector in there, compute stuff,
  # and then restore the dynamic_variables
  sD, sSr, sτ = state
  bD .= D
  bSr .= Sr
  bτ .= τ
  D .= sD
  Sr .= sSr
  τ .= sτ
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, P.equation, mesh)
  broadcast_volume!(flux_source_spherical1d, P.equation, mesh)
  D .= bD
  Sr .= bSr
  τ .= bτ

  periodic = isperiodic(mesh)
  prms_slope_limiter = (; method=P.prms.slope_limiter_method,
                          tvb_M=P.prms.slope_limiter_tvb_M)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)
  dl = L/K

  id = P.prms.id
  bc = P.prms.bc
  if id == "bondi_accretion" || id == "bondi_accretion_infall"
    if bc == "from_id"
      @unpack init_ρ, init_vr, init_ϵ, r = get_static_variables(mesh)
      dr = r[2]-r[1]
      bdry_γrr_l, bdry_γrr_ll = extrapolate_quadratic(γrr,     r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_α_l,   bdry_α_ll   = extrapolate_quadratic(α,       r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_βru_l, bdry_βru_ll = extrapolate_quadratic(βru,     r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_ρ_l,   bdry_ρ_ll   = extrapolate_quadratic(ρ,       r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_vr_l,  bdry_vr_ll  = extrapolate_quadratic(vr,      r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_ϵ_l,   bdry_ϵ_ll   = extrapolate_quadratic(ϵ,       r, 1:3,   r[1]-dr, r[1]-2*dr)
      bdry_γrr_r, bdry_γrr_rr = extrapolate_quadratic(γrr,     r, K-2:K, r[K]+dr, r[K]+2*dr)
      bdry_α_r,   bdry_α_rr   = extrapolate_quadratic(α,       r, K-2:K, r[K]+dr, r[K]+2*dr)
      bdry_βru_r, bdry_βru_rr = extrapolate_quadratic(βru,     r, K-2:K, r[K]+dr, r[K]+2*dr)
      # only outer boundary is inflow bdry
      bdry_ρ_r,   bdry_ρ_rr   = extrapolate_quadratic(init_ρ,  r, K-2:K, r[K]+dr, r[K]+2*dr)
      bdry_vr_r,  bdry_vr_rr  = extrapolate_quadratic(init_vr, r, K-2:K, r[K]+dr, r[K]+2*dr)
      bdry_ϵ_r,   bdry_ϵ_rr   = extrapolate_quadratic(init_ϵ,  r, K-2:K, r[K]+dr, r[K]+2*dr)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  elseif id == "tov"
    if bc == "tov_reflective_origin"
      bdry_γrr_l, bdry_γrr_ll = γrr[1],   γrr[2]
      bdry_α_l,   bdry_α_ll   = α[1],     α[2]
      bdry_βru_l, bdry_βru_ll = -βru[1],  -βru[2]
      bdry_ρ_l,   bdry_ρ_ll   = ρ[1],     ρ[2]
      bdry_vr_l,  bdry_vr_ll  = -vr[1],   -vr[2]
      bdry_ϵ_l,   bdry_ϵ_ll   = ϵ[1],     ϵ[2]
      bdry_γrr_r, bdry_γrr_rr = γrr[end], γrr[end]
      bdry_α_r,   bdry_α_rr   = α[end],   α[end]
      bdry_βru_r, bdry_βru_rr = βru[end], βru[end]
      bdry_ρ_r,   bdry_ρ_rr   = ρ[end],   ρ[end]
      bdry_vr_r,  bdry_vr_rr  = vr[end],  vr[end]
      bdry_ϵ_r,   bdry_ϵ_rr   = ϵ[end],   ϵ[end]
    elseif bc == "tov_symmetric_domain"
      bdry_γrr_l, bdry_γrr_ll = γrr[1],   γrr[1]
      bdry_α_l,   bdry_α_ll   = α[1],     α[1]
      bdry_βru_l, bdry_βru_ll = βru[1],   βru[1]
      bdry_ρ_l,   bdry_ρ_ll   = ρ[1],     ρ[1]
      bdry_vr_l,  bdry_vr_ll  = vr[1],    vr[1]
      bdry_ϵ_l,   bdry_ϵ_ll   = ϵ[1],     ϵ[1]
      bdry_γrr_r, bdry_γrr_rr = γrr[end], γrr[end]
      bdry_α_r,   bdry_α_rr   = α[end],   α[end]
      bdry_βru_r, bdry_βru_rr = βru[end], βru[end]
      bdry_ρ_r,   bdry_ρ_rr   = ρ[end],   ρ[end]
      bdry_vr_r,  bdry_vr_rr  = vr[end],  vr[end]
      bdry_ϵ_r,   bdry_ϵ_rr   = ϵ[end],   ϵ[end]
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  else
  end
  bγrr      = BdryBufferedData(bdry_γrr_ll, bdry_γrr_l, γrr, bdry_γrr_r, bdry_γrr_rr)
  bα        = BdryBufferedData(bdry_α_ll,   bdry_α_l,   α,   bdry_α_r,   bdry_α_rr)
  bβru      = BdryBufferedData(bdry_βru_ll, bdry_βru_l, βru, bdry_βru_r, bdry_βru_rr)
  bρ        = BdryBufferedData(bdry_ρ_ll,   bdry_ρ_l,   ρ,   bdry_ρ_r,   bdry_ρ_rr)
  bvr       = BdryBufferedData(bdry_vr_ll,  bdry_vr_l,  vr,  bdry_vr_r,  bdry_vr_rr)
  bϵ        = BdryBufferedData(bdry_ϵ_ll,   bdry_ϵ_l,   ϵ,   bdry_ϵ_r,   bdry_ϵ_rr)
  bρL       = BdryBufferedData(0, 0, ρL,  0, 0)
  bvrL      = BdryBufferedData(0, 0, vrL, 0, 0)
  bϵL       = BdryBufferedData(0, 0, ϵL,  0, 0)
  bpL       = BdryBufferedData(0, 0, pL,  0, 0)
  bρR       = BdryBufferedData(0, 0, ρR,  0, 0)
  bvrR      = BdryBufferedData(0, 0, vrR, 0, 0)
  bϵR       = BdryBufferedData(0, 0, ϵR,  0, 0)
  bpR       = BdryBufferedData(0, 0, pR,  0, 0)
  bdρ_m     = BdryBufferedData(0, 0, dρ_m, 0, 0)
  bdρ_p     = BdryBufferedData(0, 0, dρ_p, 0, 0)
  bdvr_m    = BdryBufferedData(0, 0, dvr_m, 0, 0)
  bdvr_p    = BdryBufferedData(0, 0, dvr_p, 0, 0)
  bdϵ_m     = BdryBufferedData(0, 0, dϵ_m, 0, 0)
  bdϵ_p     = BdryBufferedData(0, 0, dϵ_p, 0, 0)
  @unpack max_v = get_static_variables(mesh)
  bde_max_v = BdryBufferedData(max_v[1], max_v[1], max_v, max_v[end], max_v[end])

  for vars in ( (bρL,bρR,bρ,bdρ_m,bdρ_p),
                (bvrL,bvrR,bvr,bdvr_m,bdvr_p),
                (bϵL,bϵR,bϵ,bdϵ_m,bdϵ_p) )

    wL,wR,w,dw₋,dw₊ = vars

    # compute left (-) and right (+) differences
    for k in 0:K+1
      k₋, k₊ = k-1, k+1
      w₋, w₀, w₊ = w[k₋], w[k], w[k₊]
      dw₊[k], dw₋[k] = w₊-w₀, w₀-w₋
      dwL = limit_slope(dw₊[k], dw₋[k], dl, prms_slope_limiter)
      wL[k] = w[k] - dwL/2
      wR[k] = w[k] + dwL/2
    end
  end

  for k in 0:K+1
    bpL[k] = eos(Pressure, Density, bρL[k], InternalEnergy, bϵL[k])
    bpR[k] = eos(Pressure, Density, bρR[k], InternalEnergy, bϵR[k])
  end

  vars = (;bγrr, bα, bβru, bρ, bvr, bϵ, bρL, bvrL, bϵL, bpL,
           bρR, bvrR, bϵR, bpR, bdρ_m, bdρ_p, bdvr_m, bdvr_p, bdϵ_m, bdϵ_p, bde_max_v,
           nflx_D_L, nflx_Sr_L, nflx_τ_L, nflx_D_R, nflx_Sr_R, nflx_τ_R, eos)
  numflux_method = P.prms.fv_numerical_flux
  if numflux_method == "llf"
    numflux_local_lax_friedrich!(vars, K, P)
  else
    error("numerical flux method $(numflux_method) not implemented")
  end

  # evaluate rhs
  # we have a + in the braces, because we use the nx with fluxes above
  if :D ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_D[k]  = -(nflx_D_R[k]  + nflx_D_L[k] )/dl + src_D[k]
    end
  end
  if :Sr ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_Sr[k] = -(nflx_Sr_R[k] + nflx_Sr_L[k])/dl + src_Sr[k]
    end
  end
  if :τ ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_τ[k]  = -(nflx_τ_R[k]  + nflx_τ_L[k] )/dl + src_τ[k]
    end
  end

  # for the outer bdry conditions for the TOV star we cheat here
  if bc == "tov_reflective_origin"
    rhs_D[end] = rhs_Sr[end] = rhs_τ[end] = 0
  elseif bc == "tov_symmetric_domain"
    rhs_D[1] = rhs_Sr[1] = rhs_τ[1] = rhs_D[end] = rhs_Sr[end] = rhs_τ[end] = 0
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_spherical1d, P.equation, mesh)
  end

  return
end


@inline function numflux_local_lax_friedrich!(vars, K, P::Project{:spherical1d})
  @unpack bγrr, bα, bβru, bρ, bvr, bϵ, bρL, bvrL, bϵL, bpL,
          bρR, bvrR, bϵR, bpR, bdρ_m, bdρ_p, bdvr_m, bdvr_p, bdϵ_m, bdϵ_p, bde_max_v,
          nflx_D_L, nflx_Sr_L, nflx_τ_L, nflx_D_R, nflx_Sr_R, nflx_τ_R, eos = vars
  for k in 1:K
    # numerical fluxes
    # subscript + ... cell interior value
    # subscript - ... cell exterior value
    # nx ... pointing from interior to exterior cell
    for (k₊,k₋,nx, # cell index with interior value, cell index with exterior value, normal
         bvr₊,bρ₊,bϵ₊,bp₊, # arrays with cell interior data
         bvr₋,bρ₋,bϵ₋,bp₋, # arrays with cell exterior data
         nflx_D,nflx_Sr,nflx_τ) in (
                        # left cell interfaces
                        (k,k-1,-1.0,
                         bvrL, bρL, bϵL, bpL,
                         bvrR, bρR, bϵR, bpR,
                         vars.nflx_D_L,vars.nflx_Sr_L,vars.nflx_τ_L),
                        # right cell interfaces
                        (k,k+1, 1.0,
                         bvrR, bρR, bϵR, bpR,
                         bvrL, bρL, bϵL, bpL,
                         vars.nflx_D_R,vars.nflx_Sr_R,vars.nflx_τ_R)
                      )

      γrr₊, γrr₋     = bγrr[k₊], bγrr[k₋]
      γuurr₊, γuurr₋ = 1/γrr₊, 1/γrr₋
      vr₊, vr₋       = bvr₊[k₊], bvr₋[k₋]
      ρ₊, ρ₋         = bρ₊[k₊], bρ₋[k₋]
      ϵ₊, ϵ₋         = bϵ₊[k₊], bϵ₋[k₋]
      p₊, p₋         = bp₊[k₊], bp₋[k₋]
      vru₊, vru₋     = vr₊*γuurr₊, vr₋*γuurr₋
      W₊, W₋         = 1/sqrt(1-vru₊*vr₊), 1/sqrt(1-vru₋*vr₋)
      D₊, D₋         = ρ₊*W₊, ρ₋*W₋
      h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
      Sr₊, Sr₋       = ρ₊*h₊*W₊^2*vr₊, ρ₋*h₋*W₋^2*vr₋
      τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
      α₊, α₋         = bα[k₊], bα[k₋]
      βru₊, βru₋     = bβru[k₊], bβru[k₋]
      vv             = absmax(bde_max_v[k₊], bde_max_v[k₋])
      f₊, f₋         = α₊*D₊*(vru₊-βru₊/α₊), α₋*D₋*(vru₋-βru₋/α₋)
      nflx_D[k]      = LLF(nx*f₊, nx*f₋, D₊, D₋, vv)
      f₊, f₋         = α₊*(Sr₊*(vru₊-βru₊/α₊)+p₊), α₋*(Sr₋*(vru₋-βru₋/α₋)+p₋)
      nflx_Sr[k]     = LLF(nx*f₊, nx*f₋, Sr₊, Sr₋, vv)
      f₊, f₋         = α₊*(τ₊*(vru₊-βru₊/α₊)+p₊*vru₊), α₋*(τ₋*(vru₋-βru₋/α₋)+p₋*vru₋)
      nflx_τ[k]      = LLF(nx*f₊, nx*f₋, τ₊, τ₋, vv)
    end
  end
end


function rhs_fv_slope_limiter!(mesh, P::Project{:doublecartoon})
  @unpack equation                = P
  @unpack rD, rSx, rτ             = get_dynamic_variables(mesh)
  @unpack rD1, rSx1, rτ1,
          rD2, rSx2, rτ2          = get_static_variables(mesh)
  @unpack rhs_rD, rhs_rSx, rhs_rτ = get_rhs_variables(mesh)
  @unpack dt = get_global_variables(mesh)
  dt = dt[1]

  # SSPRK3 stepping from Hesthaven 2018 book, sec 10.2 (cf Script 10.13)
  step_fv_slope_limiter!(mesh, P, (rD,rSx,rτ))
  if :rD ∉ P.prms.freeze_vars
    @. rD1  = rD  + dt*rhs_rD
  end
  if :rSx ∉ P.prms.freeze_vars
    @. rSx1 = rSx + dt*rhs_rSx
  end
  if :rτ ∉ P.prms.freeze_vars
    @. rτ1  = rτ  + dt*rhs_rτ
  end

  step_fv_slope_limiter!(mesh, P, (rD1,rSx1,rτ1))
  if :rD ∉ P.prms.freeze_vars
    @. rD2  = (3*rD  + rD1  + dt*rhs_rD)/4
  end
  if :rSx ∉ P.prms.freeze_vars
    @. rSx2 = (3*rSx + rSx1 + dt*rhs_rSx)/4
  end
  if :rτ ∉ P.prms.freeze_vars
    @. rτ2  = (3*rτ  + rτ1  + dt*rhs_rτ)/4
  end

  step_fv_slope_limiter!(mesh, P, (rD2,rSx2,rτ2))
  # atm final result is published via rhs_ variables
  if :rD ∉ P.prms.freeze_vars
    @. rhs_rD  = (rD  + 2*rD2  + 2*dt*rhs_rD)/3
  end
  if :rSx ∉ P.prms.freeze_vars
    @. rhs_rSx = (rSx + 2*rSx2 + 2*dt*rhs_rSx)/3
  end
  if :rτ ∉ P.prms.freeze_vars
    @. rhs_rτ  = (rτ  + 2*rτ2  + 2*dt*rhs_rτ)/3
  end

  return
end


function step_fv_slope_limiter!(mesh, P::Project{:doublecartoon}, state)

  @unpack equation                      = P
  @unpack rD, rSx, rτ                   = get_dynamic_variables(mesh)
  @unpack rhs_rD, rhs_rSx, rhs_rτ       = get_rhs_variables(mesh)
  @unpack brD, brSx, brτ,
          ρ, vx, ϵ,
          ρL, vxL, ϵL, pL,
          ρR, vxR, ϵR, pR,
          dρ_p, dvx_p, dϵ_p,
          dρ_m, dvx_m, dϵ_m,
          nflx_rD_L, nflx_rSx_L, nflx_rτ_L,
          nflx_rD_R, nflx_rSx_R, nflx_rτ_R,
          src_rD, src_rSx, src_rτ,
          γxx, α, βux, sqrtdetγ         = get_static_variables(mesh)
  @unpack eos                           = equation

  # this is a hack
  # the problem is that all @with_signature methods are tied to the
  # dynamic_variables of mesh.cache (D,Sr,τ)
  # but because we don't use the ordinary evolve interface, we have
  # to handle the state vector here manually
  # this requires us to first backup the current dynamic_variables,
  # copy the new state vector in there, compute stuff,
  # and then restore the dynamic_variables
  sD, sSx, sτ = state
  brD .= rD
  brSx .= rSx
  brτ .= rτ
  rD .= sD
  rSx .= sSx
  rτ .= sτ
  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, P.equation, mesh)
  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, P.equation, mesh)
  rD .= brD
  rSx .= brSx
  rτ .= brτ

  periodic = isperiodic(mesh)
  prms_slope_limiter = (; method=P.prms.slope_limiter_method,
                          tvb_M=P.prms.slope_limiter_tvb_M)
  K, = mesh.tree.dims
  L, = dg1d.widths(mesh)
  dl = L/K

  bc = P.prms.bc
  if bc == "tov_symmetric_domain"
    bdry_γxx_l, bdry_γxx_ll           = γxx[1],        γxx[2]
    bdry_α_l, bdry_α_ll               = α[1],          α[2]
    bdry_βux_l, bdry_βux_ll           = βux[1],        βux[2]
    bdry_sqrtdetγ_l, bdry_sqrtdetγ_ll = sqrtdetγ[1],   sqrtdetγ[2]
    bdry_ρ_l, bdry_ρ_ll               = ρ[1],          ρ[2]
    bdry_vx_l, bdry_vx_ll             = vx[1],         vx[2]
    bdry_ϵ_l, bdry_ϵ_ll               = ϵ[1],          ϵ[2]
    bdry_γxx_r,      bdry_γxx_rr      = γxx[end],      γxx[end]
    bdry_α_r,        bdry_α_rr        = α[end],        α[end]
    bdry_βux_r,      bdry_βux_rr      = βux[end],      βux[end]
    bdry_sqrtdetγ_r, bdry_sqrtdetγ_rr = sqrtdetγ[end], sqrtdetγ[end]
    bdry_ρ_r,        bdry_ρ_rr        = ρ[end],        ρ[end]
    bdry_vx_r,       bdry_vx_rr       = vx[end],       vx[end]
    bdry_ϵ_r,        bdry_ϵ_rr        = ϵ[end],        ϵ[end]
  elseif bc == "tov_reflective_origin"
    bdry_γxx_l, bdry_γxx_ll           = γxx[1],        γxx[2]
    bdry_α_l, bdry_α_ll               = α[1],          α[2]
    bdry_βux_l, bdry_βux_ll           = -βux[1],       -βux[2]
    bdry_sqrtdetγ_l, bdry_sqrtdetγ_ll = sqrtdetγ[1],   sqrtdetγ[2]
    bdry_ρ_l, bdry_ρ_ll               = ρ[1],          ρ[2]
    bdry_vx_l, bdry_vx_ll             = -vx[1],        -vx[2]
    bdry_ϵ_l, bdry_ϵ_ll               = ϵ[1],          ϵ[2]
    bdry_γxx_r,      bdry_γxx_rr      = γxx[end],      γxx[end]
    bdry_α_r,        bdry_α_rr        = α[end],        α[end]
    bdry_βux_r,      bdry_βux_rr      = βux[end],      βux[end]
    bdry_sqrtdetγ_r, bdry_sqrtdetγ_rr = sqrtdetγ[end], sqrtdetγ[end]
    bdry_ρ_r,        bdry_ρ_rr        = ρ[end],        ρ[end]
    bdry_vx_r,       bdry_vx_rr       = vx[end],       vx[end]
    bdry_ϵ_r,        bdry_ϵ_rr        = ϵ[end],        ϵ[end]
  else
    TODO("implement boundary condition $bc")
  end
  bγxx      = BdryBufferedData(bdry_γxx_ll,      bdry_γxx_l,      γxx,      bdry_γxx_r,      bdry_γxx_rr)
  bα        = BdryBufferedData(bdry_α_ll,        bdry_α_l,        α,        bdry_α_r,        bdry_α_rr)
  bβux      = BdryBufferedData(bdry_βux_ll,      bdry_βux_l,      βux,      bdry_βux_r,      bdry_βux_rr)
  bsqrtdetγ = BdryBufferedData(bdry_sqrtdetγ_ll, bdry_sqrtdetγ_l, sqrtdetγ, bdry_sqrtdetγ_r, bdry_sqrtdetγ_rr)
  bρ        = BdryBufferedData(bdry_ρ_ll,        bdry_ρ_l,        ρ,        bdry_ρ_r,        bdry_ρ_rr)
  bvx       = BdryBufferedData(bdry_vx_ll,       bdry_vx_l,       vx,       bdry_vx_r,       bdry_vx_rr)
  bϵ        = BdryBufferedData(bdry_ϵ_ll,        bdry_ϵ_l,        ϵ,        bdry_ϵ_r,        bdry_ϵ_rr)
  bρL       = BdryBufferedData(0, 0, ρL,  0, 0)
  bvxL      = BdryBufferedData(0, 0, vxL, 0, 0)
  bϵL       = BdryBufferedData(0, 0, ϵL,  0, 0)
  bpL       = BdryBufferedData(0, 0, pL,  0, 0)
  bρR       = BdryBufferedData(0, 0, ρR,  0, 0)
  bvxR      = BdryBufferedData(0, 0, vxR, 0, 0)
  bϵR       = BdryBufferedData(0, 0, ϵR,  0, 0)
  bpR       = BdryBufferedData(0, 0, pR,  0, 0)
  bdρ_m     = BdryBufferedData(0, 0, dρ_m, 0, 0)
  bdρ_p     = BdryBufferedData(0, 0, dρ_p, 0, 0)
  bdvx_m    = BdryBufferedData(0, 0, dvx_m, 0, 0)
  bdvx_p    = BdryBufferedData(0, 0, dvx_p, 0, 0)
  bdϵ_m     = BdryBufferedData(0, 0, dϵ_m, 0, 0)
  bdϵ_p     = BdryBufferedData(0, 0, dϵ_p, 0, 0)
  @unpack max_v = get_static_variables(mesh)
  bde_max_v = BdryBufferedData(max_v[2], max_v[1], max_v, max_v[end], max_v[end-1])

  for vars in ( (bρL,bρR,bρ,bdρ_m,bdρ_p),
                (bvxL,bvxR,bvx,bdvx_m,bdvx_p),
                (bϵL,bϵR,bϵ,bdϵ_m,bdϵ_p) )
    wL,wR,w,dw₋,dw₊ = vars
    # compute left (-) and right (+) differences
    for k in 0:K+1
      k₋, k₊ = k-1, k+1
      w₋, w₀, w₊ = w[k₋], w[k], w[k₊]
      dw₊[k], dw₋[k] = w₊-w₀, w₀-w₋
      dwL = limit_slope(dw₊[k], dw₋[k], dl, prms_slope_limiter)
      wL[k] = w[k] - dwL/2
      wR[k] = w[k] + dwL/2
    end
  end

  for k in 0:K+1
    bpL[k] = eos(Pressure, Density, bρL[k], InternalEnergy, bϵL[k])
    bpR[k] = eos(Pressure, Density, bρR[k], InternalEnergy, bϵR[k])
  end

  vars = (;bγxx, bα, bβux, bsqrtdetγ, bρ, bvx, bϵ, bρL, bvxL, bϵL, bpL,
           bρR, bvxR, bϵR, bpR, bdρ_m, bdρ_p, bdvx_m, bdvx_p, bdϵ_m, bdϵ_p, bde_max_v,
           nflx_rD_L, nflx_rSx_L, nflx_rτ_L, nflx_rD_R, nflx_rSx_R, nflx_rτ_R, eos)
  numflux_method = P.prms.fv_numerical_flux
  if numflux_method == "llf"
    numflux_local_lax_friedrich!(vars, K, P)
  elseif numflux_method == "roe"
    numflux_roe!(vars, K, P)
  elseif numflux_method == "marquina"
    numflux_marquina!(vars, K, P)
  else
    error("numerical flux method $(numflux_method) not implemented")
  end

  # evaluate rhs
  # we have a + in the braces, because we use the nx with fluxes above
  if :rD ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_rD[k]  = -(nflx_rD_R[k]  + nflx_rD_L[k] )/dl + src_rD[k]
    end
  end
  if :rSr ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_rSx[k] = -(nflx_rSx_R[k] + nflx_rSx_L[k])/dl + src_rSx[k]
    end
  end
  if :rτ ∉ P.prms.freeze_vars
    @turbo for k in 1:K
      rhs_rτ[k]  = -(nflx_rτ_R[k]  + nflx_rτ_L[k] )/dl + src_rτ[k]
    end
  end

  # for the outer bdry conditions for the TOV star we cheat here
  if bc == "tov_reflective_origin"
    rhs_rD[end] = rhs_rSx[end] = rhs_rτ[end] = 0
  elseif bc == "tov_symmetric_domain"
    rhs_rD[1] = rhs_rSx[1] = rhs_rτ[1] = rhs_rD[end] = rhs_rSx[end] = rhs_rτ[end] = 0
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_doublecartoon, P.equation, mesh)
  end

  return
end

@inline function numflux_local_lax_friedrich!(vars, K, P::Project{:doublecartoon})
  @unpack bγxx, bα, bβux, bsqrtdetγ, bρ, bvx, bϵ, bρL, bvxL, bϵL, bpL,
          bρR, bvxR, bϵR, bpR, bdρ_m, bdρ_p, bdvx_m, bdvx_p, bdϵ_m, bdϵ_p, bde_max_v,
          nflx_rD_L, nflx_rSx_L, nflx_rτ_L, nflx_rD_R, nflx_rSx_R, nflx_rτ_R, eos = vars
  for k in 1:K
    # numerical fluxes
    # subscript + ... cell interior value
    # subscript - ... cell exterior value
    # nx ... pointing from interior to exterior cell
    for (k₊,k₋,nx, # cell index with interior value, cell index with exterior value, normal
         bvx₊,bρ₊,bϵ₊,bp₊, # arrays with cell interior data
         bvx₋,bρ₋,bϵ₋,bp₋, # arrays with cell exterior data
         nflx_rD,nflx_rSx,nflx_rτ) in (
                        # left cell interfaces
                        (k,k-1,-1.0,
                         bvxL, bρL, bϵL, bpL,
                         bvxR, bρR, bϵR, bpR,
                         vars.nflx_rD_L,vars.nflx_rSx_L,vars.nflx_rτ_L),
                        # right cell interfaces
                        (k,k+1, 1.0,
                         bvxR, bρR, bϵR, bpR,
                         bvxL, bρL, bϵL, bpL,
                         vars.nflx_rD_R,vars.nflx_rSx_R,vars.nflx_rτ_R)
                      )

      γxx₀           = (bγxx[k₊]+bγxx[k₋])/2
      α₀             = (bα[k₊]+bα[k₋])/2
      βux₀           = (bβux[k₊]+bβux[k₋])/2
      sdγ₀           = (bsqrtdetγ[k₊]+bsqrtdetγ[k₋])/2
      γuuxx₀         = 1/γxx₀
      vx₊, vx₋       = bvx₊[k₊], bvx₋[k₋]
      ρ₊, ρ₋         = bρ₊[k₊], bρ₋[k₋]
      ϵ₊, ϵ₋         = bϵ₊[k₊], bϵ₋[k₋]
      p₊, p₋         = bp₊[k₊], bp₋[k₋]
      vxu₊, vxu₋     = vx₊*γuuxx₀, vx₋*γuuxx₀
      W₊, W₋         = 1/sqrt(1-vxu₊*vx₊), 1/sqrt(1-vxu₋*vx₋)
      D₊, D₋         = ρ₊*W₊, ρ₋*W₋
      h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
      Sx₊, Sx₋       = ρ₊*h₊*W₊^2*vx₊, ρ₋*h₋*W₋^2*vx₋
      τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
      vv             = absmax(bde_max_v[k₊], bde_max_v[k₋])

      f₊, f₋         = sdγ₀*α₀*D₊*(vxu₊-βux₀/α₀), sdγ₀*α₀*D₋*(vxu₋-βux₀/α₀)
      nflx_rD[k]     = LLF(nx*f₊, nx*f₋, sdγ₀*D₊, sdγ₀*D₋, vv)
      f₊, f₋         = sdγ₀*α₀*(Sx₊*(vxu₊-βux₀/α₀)+p₊), sdγ₀*α₀*(Sx₋*(vxu₋-βux₀/α₀)+p₋)
      nflx_rSx[k]    = LLF(nx*f₊, nx*f₋, sdγ₀*Sx₊, sdγ₀*Sx₋, vv)
      f₊, f₋         = sdγ₀*α₀*(τ₊*(vxu₊-βux₀/α₀)+p₊*vxu₊), sdγ₀*α₀*(τ₋*(vxu₋-βux₀/α₀)+p₋*vxu₋)
      nflx_rτ[k]     = LLF(nx*f₊, nx*f₋, sdγ₀*τ₊, sdγ₀*τ₋, vv)
    end
  end
end
@inline function numflux_roe!(vars, K, P::Project{:doublecartoon})
  @unpack bγxx, bα, bβux, bsqrtdetγ, bρ, bvx, bϵ, bρL, bvxL, bϵL, bpL,
          bρR, bvxR, bϵR, bpR, bdρ_m, bdρ_p, bdvx_m, bdvx_p, bdϵ_m, bdϵ_p, bde_max_v,
          nflx_rD_L, nflx_rSx_L, nflx_rτ_L, nflx_rD_R, nflx_rSx_R, nflx_rτ_R, eos = vars
  for k in 1:K
    # numerical fluxes
    # subscript + ... cell interior value
    # subscript - ... cell exterior value
    # nx ... pointing from interior to exterior cell
    for (k₊,k₋,nx, # cell index with interior value, cell index with exterior value, normal
         bvx₊,bρ₊,bϵ₊,bp₊, # arrays with cell interior data
         bvx₋,bρ₋,bϵ₋,bp₋, # arrays with cell exterior data
         nflx_rD,nflx_rSx,nflx_rτ) in (
                        # left cell interfaces
                        (k,k-1,-1.0,
                         bvxL, bρL, bϵL, bpL,
                         bvxR, bρR, bϵR, bpR,
                         vars.nflx_rD_L,vars.nflx_rSx_L,vars.nflx_rτ_L),
                        # right cell interfaces
                        (k,k+1, 1.0,
                         bvxR, bρR, bϵR, bpR,
                         bvxL, bρL, bϵL, bpL,
                         vars.nflx_rD_R,vars.nflx_rSx_R,vars.nflx_rτ_R)
                      )

      γxx₀           = (bγxx[k₊]+bγxx[k₋])/2
      α₀             = (bα[k₊]+bα[k₋])/2
      βux₀           = (bβux[k₊]+bβux[k₋])/2
      sdγ₀           = (bsqrtdetγ[k₊]+bsqrtdetγ[k₋])/2
      γuuxx₀         = 1/γxx₀
      vx₊, vx₋       = bvx₊[k₊], bvx₋[k₋]
      ρ₊, ρ₋         = bρ₊[k₊], bρ₋[k₋]
      ϵ₊, ϵ₋         = bϵ₊[k₊], bϵ₋[k₋]
      p₊, p₋         = bp₊[k₊], bp₋[k₋]
      vxu₊, vxu₋     = vx₊*γuuxx₀, vx₋*γuuxx₀
      W₊, W₋         = 1/sqrt(1-vxu₊*vx₊), 1/sqrt(1-vxu₋*vx₋)
      D₊, D₋         = ρ₊*W₊, ρ₋*W₋
      h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
      Sx₊, Sx₋       = ρ₊*h₊*W₊^2*vx₊, ρ₋*h₋*W₋^2*vx₋
      τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
      vv             = absmax(bde_max_v[k₊], bde_max_v[k₋])

      vx₀ = (vx₊+vx₋)/2
      vux₀ = γuuxx₀*vx₀
      v2₀ = vux₀*vx₀
      ρ₀  = (ρ₊+ρ₋)/2
      ϵ₀  = (ϵ₊+ϵ₋)/2
      p₀  = (p₊+p₋)/2
      h₀ = 1+ϵ₀+p₀/ρ₀
      cs2 = eos(SquareSpeedOfSound, Density, ρ₀, InternalEnergy, ϵ₀)
      cs = sqrt(cs2)
      κ₀ = eos(Derivative{Pressure,InternalEnergy}, Density, ρ₀, InternalEnergy, ϵ₀)
      χ₀ = eos(Derivative{Pressure,Density}, Density, ρ₀, InternalEnergy, ϵ₀)
      κκ₀ = κ₀/ρ₀
      KK = κκ₀/(κκ₀-cs2)
      λp = α₀/(1-v2₀*cs2)*(vux₀*(1-cs2)+cs*sqrt((1-v2₀)*(γuuxx₀*(1-v2₀*cs2)-vux₀*vux₀*(1-cs2))))-βux₀
      λ0 = α₀*vux₀-βux₀
      λm = α₀/(1-v2₀*cs2)*(vux₀*(1-cs2)-cs*sqrt((1-v2₀)*(γuuxx₀*(1-v2₀*cs2)-vux₀*vux₀*(1-cs2))))-βux₀
      Λp = λp/α₀+βux₀/α₀
      Λm = λm/α₀+βux₀/α₀
      Vxp = (vux₀-Λp)/(γuuxx₀-vux₀*Λp)
      Vxm = (vux₀-Λm)/(γuuxx₀-vux₀*Λm)
      Cp = vx₀-Vxp
      Cm = vx₀-Vxm
      Wx₀ = 1-vux₀*vx₀
      W₀ = 1/sqrt(1-v2₀)
      Ap = (γuuxx₀-vux₀*vux₀)/(γuuxx₀-vux₀*Λp)
      Am = (γuuxx₀-vux₀*vux₀)/(γuuxx₀-vux₀*Λm)
      Γxx = γxx₀*γxx₀ # γxx = γyy = γzz, assuming isotropic metric
      ξ = Γxx - sdγ₀*vux₀*vux₀
      Δ = h₀^3*W₀*(KK-1)*(Cp-Cm)*ξ

      # right eigenvectors
      # taken from Ibanez et al 2001, Godunov Methods: Theory and Applications, p 489
      #   this work also contains the left eigenvectors, however, some of the auxiliary
      #   variables are not defined there, e.g. Γxy etc
      # also see Font 2000, PRD 61 044011, eq 64, but only contains right eigenvectors
      #
      # here we assume vx=vz=0 and an isotropic metric
      R = @SMatrix [ KK/(h₀*W₀)      0.0        0.0       1.0                                            1.0;
                     vx₀             0.0        0.0       h₀*W₀*(vx₀-(vux₀-Λp)/(γuuxx₀-vux₀*Λp))         h₀*W₀*(vx₀-(vux₀-Λm)/(γuuxx₀-vux₀*Λm));
                     0.0             h₀*γxx₀    0.0       0.0                                            0.0;
                     0.0             0.0        h₀*γxx₀   0.0                                            0.0;
                     1.0-KK/(h₀*W₀)  0.0        0.0       h₀*W₀*(γuuxx₀-vux₀*vux₀)/(γuuxx₀-vux₀*Λp)-1.0  h₀*W₀*(γuuxx₀-vux₀*vux₀)/(γuuxx₀-vux₀*Λm)-1.0
                   ]
      Δu = @SVector [ D₋-D₊, Sx₋-Sx₊, 0.0, 0.0, τ₋-τ₊ ]
      Δω = R \ Δu

      # roe upwinding term
      # cf. Font 2000, PRD 61 044011, eq 64
      # cf. Hesthaven 2018, Numerical methods for conservation laws,
      #   sec. 3.2 - Riemann solver
      #   sec. 6.2.1 - Roe flux
      up_u = @. abs(λ0)*Δω[1]*R[:,1] + abs(λ0)*Δω[2]*R[:,2] + abs(λ0)*Δω[3]*R[:,3] + abs(λp)*Δω[4]*R[:,4] + abs(λm)*Δω[5]*R[:,5]
      up_D, up_Sx, up_τ = up_u[1], up_u[2], up_u[5]

      f₊, f₋       = sdγ₀*α₀*D₊*(vxu₊-βux₀/α₀), sdγ₀*α₀*D₋*(vxu₋-βux₀/α₀)
      nflx_rD[k]   = nx*(f₊+f₋)/2-up_D
      f₊, f₋       = sdγ₀*α₀*(Sx₊*(vxu₊-βux₀/α₀)+p₊), sdγ₀*α₀*(Sx₋*(vxu₋-βux₀/α₀)+p₋)
      nflx_rSx[k]  = nx*(f₊+f₋)/2-up_Sx
      f₊, f₋       = sdγ₀*α₀*(τ₊*(vxu₊-βux₀/α₀)+p₊*vxu₊), sdγ₀*α₀*(τ₋*(vxu₋-βux₀/α₀)+p₋*vxu₋)
      nflx_rτ[k]   = nx*(f₊+f₋)/2-up_τ
    end
  end
end
@inline function numflux_marquina!(vars, K, P::Project{:doublecartoon})
  @unpack bγxx, bα, bβux, bsqrtdetγ, bρ, bvx, bϵ, bρL, bvxL, bϵL, bpL,
          bρR, bvxR, bϵR, bpR, bdρ_m, bdρ_p, bdvx_m, bdvx_p, bdϵ_m, bdϵ_p, bde_max_v,
          nflx_rD_L, nflx_rSx_L, nflx_rτ_L, nflx_rD_R, nflx_rSx_R, nflx_rτ_R, eos = vars
  R₊, R₋ = [ MMatrix{5,5,Float64,25}(0 for _ in 1:25) for _ in 1:2 ]
  for k in 1:K
    # numerical fluxes
    # subscript + ... cell interior value
    # subscript - ... cell exterior value
    # nx ... pointing from interior to exterior cell
    for (k₊,k₋,nx, # cell index with interior value, cell index with exterior value, normal
         bvx₊,bρ₊,bϵ₊,bp₊, # arrays with cell interior data
         bvx₋,bρ₋,bϵ₋,bp₋, # arrays with cell exterior data
         nflx_rD,nflx_rSx,nflx_rτ) in (
                        # left cell interfaces
                        (k,k-1,-1.0,
                         bvxL, bρL, bϵL, bpL,
                         bvxR, bρR, bϵR, bpR,
                         vars.nflx_rD_L,vars.nflx_rSx_L,vars.nflx_rτ_L),
                        # right cell interfaces
                        (k,k+1, 1.0,
                         bvxR, bρR, bϵR, bpR,
                         bvxL, bρL, bϵL, bpL,
                         vars.nflx_rD_R,vars.nflx_rSx_R,vars.nflx_rτ_R)
                      )

      γxx₀           = (bγxx[k₊]+bγxx[k₋])/2
      α₀             = (bα[k₊]+bα[k₋])/2
      βux₀           = (bβux[k₊]+bβux[k₋])/2
      sdγ₀           = (bsqrtdetγ[k₊]+bsqrtdetγ[k₋])/2
      γuuxx₀         = 1/γxx₀
      vx₊, vx₋       = bvx₊[k₊], bvx₋[k₋]
      ρ₊, ρ₋         = bρ₊[k₊], bρ₋[k₋]
      ϵ₊, ϵ₋         = bϵ₊[k₊], bϵ₋[k₋]
      p₊, p₋         = bp₊[k₊], bp₋[k₋]
      vux₊, vux₋     = vx₊*γuuxx₀, vx₋*γuuxx₀
      W₊, W₋         = 1/sqrt(1-vux₊*vx₊), 1/sqrt(1-vux₋*vx₋)
      D₊, D₋         = ρ₊*W₊, ρ₋*W₋
      h₊, h₋         = 1+ϵ₊+p₊/ρ₊, 1+ϵ₋+p₋/ρ₋
      Sx₊, Sx₋       = ρ₊*h₊*W₊^2*vx₊, ρ₋*h₋*W₋^2*vx₋
      τ₊, τ₋         = W₊^2*ρ₊*h₊-p₊-D₊, W₋^2*ρ₋*h₋-p₋-D₋
      vv             = absmax(bde_max_v[k₊], bde_max_v[k₋])

      v2₊ = vux₊*vx₊
      cs2₊ = eos(SquareSpeedOfSound, Density, ρ₊, InternalEnergy, ϵ₊)
      # the abs here is cheating; ρ₊, ϵ₋ might have become negative due to limiting,
      # so they should really have been atmosphered before
      cs₊ = sqrt(abs(cs2₊))
      κ₊ = eos(Derivative{Pressure,InternalEnergy}, Density, ρ₊, InternalEnergy, ϵ₊)
      χ₊ = eos(Derivative{Pressure,Density}, Density, ρ₊, InternalEnergy, ϵ₊)
      κκ₊ = κ₊/ρ₊
      KK₊ = κκ₊/(κκ₊-cs2₊)
      λp₊ = α₀/(1-v2₊*cs2₊)*(vux₊*(1-cs2₊)+cs₊*sqrt((1-v2₊)*(γuuxx₀*(1-v2₊*cs2₊)-vux₊*vux₊*(1-cs2₊))))-βux₀
      λ0₊ = α₀*vux₊-βux₀
      λm₊ = α₀/(1-v2₊*cs2₊)*(vux₊*(1-cs2₊)-cs₊*sqrt((1-v2₊)*(γuuxx₀*(1-v2₊*cs2₊)-vux₊*vux₊*(1-cs2₊))))-βux₀
      Λp₊ = λp₊/α₀+βux₀/α₀
      Λm₊ = λm₊/α₀+βux₀/α₀
      Vxp₊ = (vux₊-Λp₊)/(γuuxx₀-vux₊*Λp₊)
      Vxm₊ = (vux₊-Λm₊)/(γuuxx₀-vux₊*Λm₊)
      Cp₊ = vx₊-Vxp₊
      Cm₊ = vx₊-Vxm₊
      Wx₊ = 1-vux₊*vx₊
      W₊ = 1/sqrt(1-v2₊)
      Ap₊ = (γuuxx₀-vux₊*vux₊)/(γuuxx₀-vux₊*Λp₊)
      Am₊ = (γuuxx₀-vux₊*vux₊)/(γuuxx₀-vux₊*Λm₊)
      Γxx = γxx₀*γxx₀ # γxx = γyy = γzz, assuming isotropic metric
      ξ₊ = Γxx - sdγ₀*vux₊*vux₊
      Δ₊ = h₊^3*W₊*(KK₊-1)*(Cp₊-Cm₊)*ξ₊

      # right eigenvectors
      # taken from Ibanez et al 2001, Godunov Methods: Theory and Applications, p 489
      #   this work also contains the left eigenvectors, however, some of the auxiliary
      #   variables are not defined there, e.g. Γxy etc
      # also see Font 2000, PRD 61 044011, eq 64, but only contains right eigenvectors
      #
      # here we assume vx=vz=0 and an isotropic metric
      R₊ .= @SMatrix [ KK₊/(h₊*W₊)      0.0        0.0       1.0                                             1.0;
                      vx₊              0.0        0.0       h₊*W₊*(vx₊-(vux₊-Λp₊)/(γuuxx₀-vux₊*Λp₊))        h₊*W₊*(vx₊-(vux₊-Λm₊)/(γuuxx₀-vux₊*Λm₊));
                      0.0              h₊*γxx₀    0.0       0.0                                             0.0;
                      0.0              0.0        h₊*γxx₀   0.0                                             0.0;
                      1.0-KK₊/(h₊*W₊)  0.0        0.0       h₊*W₊*(γuuxx₀-vux₊*vux₊)/(γuuxx₀-vux₊*Λp₊)-1.0  h₊*W₊*(γuuxx₀-vux₊*vux₊)/(γuuxx₀-vux₊*Λm₊)-1.0
                   ]

      v2₋ = vux₋*vx₋
      cs2₋ = eos(SquareSpeedOfSound, Density, ρ₋, InternalEnergy, ϵ₋)
      # the abs here is cheating; ρ₊, ϵ₋ might have become negative due to limiting,
      # so they should really have been atmosphered before
      cs₋ = sqrt(abs(cs2₋))
      κ₋ = eos(Derivative{Pressure,InternalEnergy}, Density, ρ₋, InternalEnergy, ϵ₋)
      χ₋ = eos(Derivative{Pressure,Density}, Density, ρ₋, InternalEnergy, ϵ₋)
      κκ₋ = κ₋/ρ₋
      KK₋ = κκ₋/(κκ₋-cs2₋)
      λp₋ = α₀/(1-v2₋*cs2₋)*(vux₋*(1-cs2₋)+cs₋*sqrt((1-v2₋)*(γuuxx₀*(1-v2₋*cs2₋)-vux₋*vux₋*(1-cs2₋))))-βux₀
      λ0₋ = α₀*vux₋-βux₀
      λm₋ = α₀/(1-v2₋*cs2₋)*(vux₋*(1-cs2₋)-cs₋*sqrt((1-v2₋)*(γuuxx₀*(1-v2₋*cs2₋)-vux₋*vux₋*(1-cs2₋))))-βux₀
      Λp₋ = λp₋/α₀+βux₀/α₀
      Λm₋ = λm₋/α₀+βux₀/α₀
      Vxp₋ = (vux₋-Λp₋)/(γuuxx₀-vux₋*Λp₋)
      Vxm₋ = (vux₋-Λm₋)/(γuuxx₀-vux₋*Λm₋)
      Cp₋ = vx₋-Vxp₋
      Cm₋ = vx₋-Vxm₋
      Wx₋ = 1-vux₋*vx₋
      W₋ = 1/sqrt(1-v2₋)
      Ap₋ = (γuuxx₀-vux₋*vux₋)/(γuuxx₀-vux₋*Λp₋)
      Am₋ = (γuuxx₀-vux₋*vux₋)/(γuuxx₀-vux₋*Λm₋)
      Γxx = γxx₀*γxx₀ # γxx = γyy = γzz, assuming isotropic metric
      ξ₋ = Γxx - sdγ₀*vux₋*vux₋
      Δ₋ = h₋^3*W₋*(KK₋-1)*(Cp₋-Cm₋)*ξ₋

      # right eigenvectors
      # taken from Ibanez et al 2001, Godunov Methods: Theory and Applications, p 489
      #   this work also contains the left eigenvectors, however, some of the auxiliary
      #   variables are not defined there, e.g. Γxy etc
      # also see Font 2000, PRD 61 044011, eq 64, but only contains right eigenvectors
      #
      # here we assume vx=vz=0 and an isotropic metric
      R₋ .= @SMatrix [ KK₋/(h₋*W₋)      0.0        0.0       1.0                                             1.0;
                      vx₋              0.0        0.0       h₋*W₋*(vx₋-(vux₋-Λp₋)/(γuuxx₀-vux₋*Λp₋))        h₋*W₋*(vx₋-(vux₋-Λm₋)/(γuuxx₀-vux₋*Λm₋));
                      0.0              h₋*γxx₀    0.0       0.0                                             0.0;
                      0.0              0.0        h₋*γxx₀   0.0                                             0.0;
                      1.0-KK₋/(h₋*W₋)  0.0        0.0       h₋*W₋*(γuuxx₀-vux₋*vux₋)/(γuuxx₀-vux₋*Λp₋)-1.0  h₋*W₋*(γuuxx₀-vux₋*vux₋)/(γuuxx₀-vux₋*Λm₋)-1.0
                   ]

      # normalize right eigenvectors
      @views R₊[:,1] ./= norm(R₊[:,1])
      @views R₊[:,2] ./= norm(R₊[:,2])
      @views R₊[:,3] ./= norm(R₊[:,3])
      @views R₊[:,4] ./= norm(R₊[:,4])
      @views R₊[:,5] ./= norm(R₊[:,5])
      @views R₋[:,1] ./= norm(R₋[:,1])
      @views R₋[:,2] ./= norm(R₋[:,2])
      @views R₋[:,3] ./= norm(R₋[:,3])
      @views R₋[:,4] ./= norm(R₋[:,4])
      @views R₋[:,5] ./= norm(R₋[:,5])

      u₊ = @SVector [ sdγ₀*D₊, sdγ₀*Sx₊, 0.0, 0.0, sdγ₀*τ₊ ]
      u₋ = @SVector [ sdγ₀*D₋, sdγ₀*Sx₋, 0.0, 0.0, sdγ₀*τ₋ ]

      frD₊, frD₋   = sdγ₀*α₀*D₊*(vux₊-βux₀/α₀),           sdγ₀*α₀*D₋*(vux₋-βux₀/α₀)
      frSx₊, frSx₋ = sdγ₀*α₀*(Sx₊*(vux₊-βux₀/α₀)+p₊),     sdγ₀*α₀*(Sx₋*(vux₋-βux₀/α₀)+p₋)
      frτ₊, frτ₋   = sdγ₀*α₀*(τ₊*(vux₊-βux₀/α₀)+p₊*vux₊), sdγ₀*α₀*(τ₋*(vux₋-βux₀/α₀)+p₋*vux₋)
      f₊ = @SVector [ frD₊, frSx₊, 0.0, 0.0, frτ₊ ]
      f₋ = @SVector [ frD₋, frSx₋, 0.0, 0.0, frτ₋ ]

      ω₊ = R₊\u₊
      ω₋ = R₋\u₋
      ϕ₊ = R₊\f₊
      ϕ₋ = R₋\f₋

      for (idx,nflx) in ( (1,nflx_rD), (2,nflx_rSx), (3,nflx_rτ) )
        tmp_nflx = 0.0
        if sign(λ0₊) == sign(λ0₋)
          if nx*λ0₊ > 0
            tmp_nflx += (ϕ₊[1]*R₊[idx,1] + ϕ₊[2]*R₊[idx,2] + ϕ₊[3]*R₊[idx,3])
          else
            tmp_nflx += (ϕ₋[1]*R₋[idx,1] + ϕ₋[2]*R₋[idx,2] + ϕ₋[3]*R₋[idx,3])
          end
        else
          vv = nx*absmax(λ0₊,λ0₋)
          tmp_nflx += ( (ϕ₊[1]+vv*ω₊[1])/2*R₊[idx,1] + (ϕ₋[1]-vv*ω₋[1])/2*R₋[idx,1] +
                        (ϕ₊[2]+vv*ω₊[2])/2*R₊[idx,2] + (ϕ₋[2]-vv*ω₋[2])/2*R₋[idx,2] +
                        (ϕ₊[3]+vv*ω₊[3])/2*R₊[idx,3] + (ϕ₋[3]-vv*ω₋[3])/2*R₋[idx,3] )
        end
        if sign(λp₊) == sign(λp₋)
          if nx*λp₊ > 0
            tmp_nflx += ϕ₊[4]*R₊[idx,4]
          else
            tmp_nflx += ϕ₋[4]*R₋[idx,4]
          end
        else
          vv = nx*absmax(λp₊,λp₋)
          tmp_nflx += (ϕ₊[4]+vv*ω₊[4])/2*R₊[idx,4] + (ϕ₋[4]-vv*ω₋[4])/2*R₋[idx,4]
        end
        if sign(λm₊) == sign(λm₋)
          if nx*λm₊ > 0
            tmp_nflx += ϕ₊[5]*R₊[idx,5]
          else
            tmp_nflx += ϕ₋[5]*R₋[idx,5]
          end
        else
          vv = nx*absmax(λm₊,λm₋)
          tmp_nflx += (ϕ₊[5]+vv*ω₊[5])/2*R₊[idx,5] + (ϕ₋[5]-vv*ω₋[5])/2*R₋[idx,5]
        end
        nflx[k] = nx*tmp_nflx
      end
    end
  end
end


@inline function limit_slope(a, b, dx, prms)
  if prms.method === :none
    return 0.0
  elseif prms.method === :minmod
    return TCI.minmod(a, b)
  elseif prms.method === :muscl
    return TCI.minmod((a+b)/2, 2*a, 2*b)
  elseif prms.method === :superbee
    return TCI.minmod(TCI.maxmod(a,b),TCI.minmod(2*a,2*b))
  elseif prms.method === :van_albada
    c = dx^3
    return TCI.minmod( ((a^2+c^2)*b+(b^2+c^2)*a)/(a^2+b^2+c^2), 2*a, 2*b )
  elseif prms.method === :van_leer
    return TCI.minmod(2*a*b/(a+b), 2*a, 2*b)
  elseif prms.method === :tvb
    return TCI.minmod_M(dx, prms.tvb_M, (a+b)/2, 2*a, 2*b)
  elseif prms.method === :mc
    return dg1d.MC(b, a)
  else
    TODO(prms.method)
  end
end


#######################################################################
#                               DG RHS                                #
#######################################################################


rhs!(mesh::Mesh1d, P::Project) = rhs!(mesh, P, P.hrsc)


function rhs!(mesh::Mesh1d, P::Project{:valencia1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                 = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ     = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          D, Sr, τ, max_v, vr, p,
          src_rD, src_rSr, src_rτ     = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_valencia1d, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_valencia1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_valencia1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    if ismirrored(mesh)
      broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
    else
      impose_symmetry_bdry_values_valencia1d!(mesh)
      broadcast_bdry!(bdryllf_valencia1d_tov_reflected, equation, mesh)
    end
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_valencia1d, P.equation, mesh)
  end

  return
end


function rhs!(mesh::Mesh1d, P::Project{:spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                    = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ        = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, vr, p,
          src_D, src_Sr, src_τ,
          freeze_mask, c2p_reset_atm  = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_c2p_reset_atm          = get_bdry_variables(cache)

  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)
  dg1d.interpolate_face_data!(mesh, c2p_reset_atm, bdry_c2p_reset_atm)

  broadcast_volume!(determine_atmosphere, P.equation, mesh)
  if P.prms.atm_equalize_on_interface
    broadcast_faces!(equalize_atmosphere_spherical1d, P.equation, mesh)
  end
  broadcast_volume!(determine_freeze_mask, P.equation, mesh)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_spherical1d, equation, mesh)
  id = P.prms.id
  bc = P.prms.bc
  if id == "bondi_accretion" || id == "bondi_accretion_infall"
    if bc == "from_id"
      # already accounts for inflow boundary
      broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  elseif id == "tov"
    if bc == "tov_symmetric_domain"
      # nothing todo, atmosphering should take care of outer bdrys
    elseif bc == "tov_reflective_origin"
      broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  else
    TODO("unknown boundary conditions for initial data $id: $bc")
  end

  if :D ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
    compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh, P.prms.bernstein, freeze_mask)
  end
  if :Sr ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
    compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh, P.prms.bernstein, freeze_mask)
  end
  if :τ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)
    compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh, P.prms.bernstein, freeze_mask)
  end

  if !P.prms.atm_evolve
    # broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_spherical1d, P.equation, mesh)
    # for k in 1:K
    #   cell = mesh.tree.cells[k]
    #   cidxs = cellindices(mesh, k)
    #   v_mask = view(mask, cidxs)
    #   s_mask = sum(v_mask)
    #   if s_mask ≈ Npts || s_mask ≈ 0.0
    #     continue
    #   end
    #   v_rhs_D  = view(rhs_D, cidxs)
    #   v_rhs_Sr = view(rhs_Sr, cidxs)
    #   v_rhs_τ  = view(rhs_τ, cidxs)
    #   if dg1d.has_neighbor(cell,Cart1d.Xmin)
    #     v_ngb_mask = dg1d.view_neighbor(mask, cell, Cart1d.Xmin, mesh)
    #     s_ngb_mask = sum(v_ngb_mask)
    #     if !(s_ngb_mask ≈ 0.0)
    #
    #     end
    #   end
    # end
  end

  if bc == "tov_symmetric_domain"
    # Need to symmetrize data around origin to avoid asymmetries to blow up there.
    # This is needed, because near the surface the solution becomes asymmetric,
    # and those errors are amplified when propagating inwards.
    # Not sure if this is caused by the LDG/AV method or by the DG method itself.
    korigin = find_cell_index(0.0, mesh)
    for (k,(v_rhs_D, v_rhs_Sr, v_rhs_τ)) in enumerate(zip(eachcell(mesh,rhs_D),
                                                          eachcell(mesh,rhs_Sr),
                                                          eachcell(mesh,rhs_τ)))
      if korigin == k
        @views v_rhs_D  .= (v_rhs_D .+ v_rhs_D[end:-1:1])./2
        @views v_rhs_Sr .= (v_rhs_Sr .- v_rhs_Sr[end:-1:1])./2
        @views v_rhs_τ  .= (v_rhs_τ .+ v_rhs_τ[end:-1:1])./2
        break
      end
    end
  end

  return
end


function rhs!(mesh::Mesh1d, P::Project{:rescaled_spherical1d}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ                    = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ        = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          max_v, vr, p, D, Sr, τ,
          src_rD, src_rSr, src_rτ        = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_rescaled_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_rescaled_spherical1d, equation, mesh)
  broadcast_faces!(llf_rescaled_spherical1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_rescaled_spherical1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_rescaled_spherical1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_rescaled_spherical1d, P.equation, mesh)
  end

  return
end


function rhs!(mesh::Mesh1d, P::Project{:doublecartoon}, hrsc::Nothing)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rτ                 = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rτ     = get_rhs_variables(cache)
  @unpack flx_rDx, flx_rSxx, flx_rτx,
          D, Sx, τ, max_v, vx, p,
          src_rD, src_rSx, src_rτ     = get_static_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rτ,
          bdry_D, bdry_Sx, bdry_τ,
          bdry_max_v, bdry_vx, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, equation, mesh)
  broadcast_faces!(llf_doublecartoon, equation, mesh)
  if ismirrored(mesh)
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi, equation, mesh)
    elseif P.prms.problem === :tov
      broadcast_bdry!(bdryllf_doublecartoon_tov, equation, mesh)
    else
      TODO()
    end
  else
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi_mirrored, equation, mesh)
    elseif P.prms.problem === :tov
      impose_symmetry_bdry_values_doublecartoon!(mesh)
      broadcast_bdry!(bdryllf_doublecartoon_tov_mirrored, equation, mesh)
    else
      TODO()
    end
  end

  if :rD ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rD,  flx_rDx,  src_rD,  nflx_rD,  mesh)
  end
  if :rSx ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rSx, flx_rSxx, src_rSx, nflx_rSx, mesh)
  end
  if :rτ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  src_rτ,  nflx_rτ,  mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_doublecartoon, P.equation, mesh)
  end

  return
end


function rhs!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractReconstruction)

  @unpack equation = P

  @unpack D, Sr, τ                    = get_dynamic_variables(mesh)
  @unpack rhs_D, rhs_Sr, rhs_τ        = get_rhs_variables(mesh)
  @unpack flr_D, flr_Sr, flr_τ,
          max_v, vr, p,
          src_D, src_Sr, src_τ        = get_static_variables(mesh)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p = get_bdry_variables(mesh)
  @unpack flag                        = get_cell_variables(mesh)

  ismirrored(mesh) || TODO("only tested for mirrored domain")

  HRSC.reconstruct!(D,  flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  HRSC.reconstruct!(Sr, flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  HRSC.reconstruct!(τ,  flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)

  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_spherical1d, equation, mesh)
  id = P.prms.id
  bc = P.prms.bc
  if id == "bondi_accretion" || id == "bondi_accretion_infall"
    if bc == "from_id"
      # already accounts for inflow boundary
      broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  elseif id == "tov"
    if bc == "tov_symmetric_domain"
      # nothing todo, atmosphering should take care of outer bdrys
    elseif bc == "tov_reflective_origin"
      broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  else
    TODO("unknown boundary conditions for initial data $id: $bc")
  end

  if :D ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  end
  if :Sr ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  end
  if :τ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_spherical1d, P.equation, mesh)
  end

  if bc == "tov_symmetric_domain"
    # Need to symmetrize data around origin to avoid asymmetries to blow up there.
    # This is needed, because near the surface the solution becomes asymmetric,
    # and those errors are amplified when propagating inwards.
    # Not sure if this is caused by the LDG/AV method or by the DG method itself.
    korigin = find_cell_index(0.0, mesh)
    for (k,(v_rhs_D, v_rhs_Sr, v_rhs_τ)) in enumerate(zip(eachcell(mesh,rhs_D),
                                                          eachcell(mesh,rhs_Sr),
                                                          eachcell(mesh,rhs_τ)))
      if korigin == k
        @views v_rhs_D  .= (v_rhs_D .+ v_rhs_D[end:-1:1])./2
        @views v_rhs_Sr .= (v_rhs_Sr .- v_rhs_Sr[end:-1:1])./2
        @views v_rhs_τ  .= (v_rhs_τ .+ v_rhs_τ[end:-1:1])./2
        break
      end
    end
  end

  return
end

function rhs!(mesh, P::Project{:doublecartoon}, hrsc::HRSC.AbstractReconstruction)

  @unpack equation = P
  @unpack rD, rSx, rτ                 = get_dynamic_variables(mesh)
  @unpack rhs_rD, rhs_rSx, rhs_rτ     = get_rhs_variables(mesh)
  @unpack flx_rDx, flx_rSxx, flx_rτx,
          D, Sx, τ, max_v, vx, p,
          src_rD, src_rSx, src_rτ     = get_static_variables(mesh)
  @unpack nflx_rD, nflx_rSx, nflx_rτ,
          bdry_D, bdry_Sx, bdry_τ,
          bdry_max_v, bdry_vx, bdry_p = get_bdry_variables(mesh)
  @unpack flag                        = get_cell_variables(mesh)

  # @unpack log_abs_rD = get_static_variables(mesh)
  # @. log_abs_rD = log(abs(max(rD,-20.0)))
  # HRSC.reconstruct!(log_abs_rD, flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=true)
  # @. rD = exp(log_abs_rD)
  HRSC.reconstruct!(rD, flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=true)
  # HRSC.reconstruct!(rSx, flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)
  # HRSC.reconstruct!(rτ,  flag, hrsc, isperiodic=isperiodic(mesh), limit_with_boundaries=false)

  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, equation, mesh)
  broadcast_faces!(llf_doublecartoon, equation, mesh)
  if ismirrored(mesh)
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi, equation, mesh)
    elseif P.prms.problem === :tov
      broadcast_bdry!(bdryllf_doublecartoon_tov, equation, mesh)
    else
      TODO()
    end
  else
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi_mirrored, equation, mesh)
    elseif P.prms.problem === :tov
      impose_symmetry_bdry_values_doublecartoon!(mesh)
      broadcast_bdry!(bdryllf_doublecartoon_tov_mirrored, equation, mesh)
    else
      TODO()
    end
  end

  if :rD ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rD,  flx_rDx,  src_rD,  nflx_rD,  mesh)
  end
  if :rSx ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rSx, flx_rSxx, src_rSx, nflx_rSx, mesh)
  end
  if :rτ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  src_rτ,  nflx_rτ,  mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_doublecartoon, P.equation, mesh)
  end

  return
end


function rhs!(mesh, P::Project{:valencia1d}, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSr, rτ             = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSr, rhs_rτ = get_rhs_variables(cache)
  @unpack flr_rD, flr_rSr, flr_rτ,
          ldg_rD, ldg_rSr, ldg_rτ,
          D, Sr, τ,
          max_v, vr, p, smoothed_mu,
          src_rD, src_rSr, src_rτ = get_static_variables(cache)
  @unpack nflr_rD, nflr_rSr, nflr_rτ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_ldg_rD, bdry_ldg_rSr, bdry_ldg_rτ,
          bdry_rhs_rD, bdry_rhs_rSr, bdry_rhs_rτ,
          bdry_smoothed_mu        = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_valencia1d, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_valencia1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  broadcast_volume!(flux_source_valencia1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
  else
    TODO()
  end

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_valencia1d, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_valencia1d, equation, mesh)
  compute_rhs_weak_form!(ldg_rD,  rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(ldg_rSr, rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(ldg_rτ,  rτ,  nflr_rτ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_{rD,rSr,rtau}
  # numerical fluxes
  broadcast_faces!(llf_valencia1d, equation, mesh)
  if P.prms.problem === :bondi
    broadcast_bdry!(bdryllf_valencia1d_bondi, equation, mesh)
  elseif P.prms.problem === :tov
    broadcast_bdry!(bdryllf_valencia1d_tov, equation, mesh)
  else
    TODO()
  end

  broadcast_volume!(av_flux_covariant_valencia1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rD,       bdry_ldg_rD)
  dg1d.interpolate_face_data!(mesh, ldg_rSr,      bdry_ldg_rSr)
  dg1d.interpolate_face_data!(mesh, ldg_rτ,       bdry_ldg_rτ)
  dg1d.interpolate_face_data!(mesh, rhs_rD,       bdry_rhs_rD)
  dg1d.interpolate_face_data!(mesh, rhs_rSr,      bdry_rhs_rSr)
  dg1d.interpolate_face_data!(mesh, rhs_rτ,       bdry_rhs_rτ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_valencia1d, equation, mesh)

  compute_rhs_weak_form!(rhs_rD,  flr_rD,  src_rD,  nflr_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSr, flr_rSr, src_rSr, nflr_rSr, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flr_rτ,  src_rτ,  nflr_rτ,  mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_valencia1d, P.equation, mesh)
  end

  return
end


function rhs!(mesh, P::Project{:spherical1d}, hrsc::HRSC.AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack D, Sr, τ                            = get_dynamic_variables(cache)
  @unpack rhs_D, rhs_Sr, rhs_τ                = get_rhs_variables(cache)
  @unpack flr_D, flr_Sr, flr_τ,
          ldg_D, ldg_Sr, ldg_τ,
          max_v, vr, p, smoothed_mu,
          src_D, src_Sr, src_τ,
          c2p_reset_atm                       = get_static_variables(cache)
  @unpack nflr_D, nflr_Sr, nflr_τ,
          bdry_D, bdry_Sr, bdry_τ,
          bdry_max_v, bdry_vr, bdry_p,
          bdry_ldg_D, bdry_ldg_Sr, bdry_ldg_τ,
          bdry_rhs_D, bdry_rhs_Sr, bdry_rhs_τ,
          bdry_smoothed_mu,
          bdry_c2p_reset_atm,                 = get_bdry_variables(cache)

  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_spherical1d_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  end
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sr,    bdry_Sr)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vr,    bdry_vr)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  if P.prms.atm_equalize_on_interface
    dg1d.interpolate_face_data!(mesh, c2p_reset_atm, bdry_c2p_reset_atm)
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_faces!(equalize_atmosphere_spherical1d, P.equation, mesh)
  end

  broadcast_volume!(flux_source_spherical1d, equation, mesh)
  impose_symmetry_sources!(P, mesh)
  broadcast_faces!(llf_spherical1d, equation, mesh)
  id = P.prms.id
  bc = P.prms.bc
  if id == "bondi_accretion" || id == "bondi_accretion_infall"
    if bc == "from_id"
      # already accounts for inflow boundary
      broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  elseif id == "tov"
    if bc == "tov_symmetric_domain"
      # nothing todo, atmosphering should take care of outer bdrys
    elseif bc == "tov_reflective_origin"
      broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  else
    TODO("unknown boundary conditions for initial data $id: $bc")
  end

  if :D ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  end
  if :Sr ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  end
  if :τ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)
  end

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_spherical1d, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_spherical1d, equation, mesh)
  compute_rhs_weak_form!(ldg_D,  D,  nflr_D,  mesh)
  compute_rhs_weak_form!(ldg_Sr, Sr, nflr_Sr, mesh)
  compute_rhs_weak_form!(ldg_τ,  τ,  nflr_τ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_{D,Sr,tau}
  # numerical fluxes
  broadcast_faces!(llf_spherical1d, equation, mesh)
  if id == "bondi_accretion" || id == "bondi_accretion_infall"
    if bc == "from_id"
      broadcast_bdry!(bdryllf_spherical1d_bondi, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  elseif id == "tov"
    if bc == "tov_symmetric_domain"
      # nothing todo, atmosphering should take care of outer bdrys
    elseif bc == "tov_reflective_origin"
      broadcast_bdry!(bdryllf_spherical1d_tov, equation, mesh)
    else
      TODO("unknown boundary conditions for initial data $id: $bc")
    end
  else
    TODO("unknown boundary conditions for initial data $id: $bc")
  end

  broadcast_volume!(av_flux_covariant_spherical1d, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_D,       bdry_ldg_D)
  dg1d.interpolate_face_data!(mesh, ldg_Sr,      bdry_ldg_Sr)
  dg1d.interpolate_face_data!(mesh, ldg_τ,       bdry_ldg_τ)
  dg1d.interpolate_face_data!(mesh, rhs_D,       bdry_rhs_D)
  dg1d.interpolate_face_data!(mesh, rhs_Sr,      bdry_rhs_Sr)
  dg1d.interpolate_face_data!(mesh, rhs_τ,       bdry_rhs_τ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu, bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_spherical1d, equation, mesh)

  if :D ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_D,  flr_D,  src_D,  nflr_D,  mesh)
  end
  if :Sr ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_Sr, flr_Sr, src_Sr, nflr_Sr, mesh)
  end
  if :τ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_τ,  flr_τ,  src_τ,  nflr_τ,  mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_spherical1d, P.equation, mesh)
  end

  if bc == "tov_symmetric_domain"
    # Need to symmetrize data around origin to avoid asymmetries to blow up there.
    # This is needed, because near the surface the solution becomes asymmetric,
    # and those errors are amplified when propagating inwards.
    # Not sure if this is caused by the LDG/AV method or by the DG method itself.
    korigin = find_cell_index(0.0, mesh)
    for (k,(v_rhs_D, v_rhs_Sr, v_rhs_τ)) in enumerate(zip(eachcell(mesh,rhs_D),
                                                          eachcell(mesh,rhs_Sr),
                                                          eachcell(mesh,rhs_τ)))
      if korigin == k
        @views v_rhs_D  .= (v_rhs_D .+ v_rhs_D[end:-1:1])./2
        @views v_rhs_Sr .= (v_rhs_Sr .- v_rhs_Sr[end:-1:1])./2
        @views v_rhs_τ  .= (v_rhs_τ .+ v_rhs_τ[end:-1:1])./2
        break
      end
    end
  end


  return
end


function rhs!(mesh::Mesh1d, P::Project{:doublecartoon}, hrsc::AbstractArtificialViscosity)

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rτ                 = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rτ     = get_rhs_variables(cache)
  @unpack flx_rDx, flx_rSxx, flx_rτx,
          ldg_rD, ldg_rSx, ldg_rτ,
          D, Sx, τ, max_v, vx, p, smoothed_mu,
          src_rD, src_rSx, src_rτ     = get_static_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rτ,
          bdry_D, bdry_Sx, bdry_τ,
          bdry_max_v, bdry_vx, bdry_p, bdry_smoothed_mu,
          bdry_ldg_rD, bdry_ldg_rSx, bdry_ldg_rτ,
          bdry_rhs_rD, bdry_rhs_rSx, bdry_rhs_rτ = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_doublecartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_doublecartoon!(P, mesh)
  broadcast_volume!(flux_source_doublecartoon, equation, mesh)
  broadcast_faces!(llf_doublecartoon, equation, mesh)
  if ismirrored(mesh)
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi, equation, mesh)
    elseif P.prms.problem === :tov
      broadcast_bdry!(bdryllf_doublecartoon_tov, equation, mesh)
    else
      TODO()
    end
  else
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi_mirrored, equation, mesh)
    elseif P.prms.problem === :tov
      impose_symmetry_bdry_values_doublecartoon!(mesh)
      broadcast_bdry!(bdryllf_doublecartoon_tov_mirrored, equation, mesh)
    else
      TODO()
    end
  end

  if :rD ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rD,  flx_rDx,  src_rD,  nflx_rD,  mesh)
  end
  if :rSx ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rSx, flx_rSxx, src_rSx, nflx_rSx, mesh)
  end
  if :rτ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  src_rτ,  nflx_rτ,  mesh)
  end

  # now compute rhs for ∂t u + ∂x f(x) = ∂x( μ (-∂t+∂x) u )

  broadcast_faces!(ldg_nflux_doublecartoon, equation, mesh)
  broadcast_bdry!(ldg_bdryllf_doublecartoon, equation, mesh)
  compute_rhs_weak_form!(ldg_rD,  rD,  nflx_rD,  mesh)
  compute_rhs_weak_form!(ldg_rSx, rSx, nflx_rSx, mesh)
  compute_rhs_weak_form!(ldg_rτ,  rτ,  nflx_rτ,  mesh)

  # need to recompute, because ldg computation overwrites nflx_r{D,Sx,tau}
  # numerical fluxes
  broadcast_faces!(llf_doublecartoon, equation, mesh)
  if ismirrored(mesh)
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi, equation, mesh)
    elseif P.prms.problem === :tov
      broadcast_bdry!(bdryllf_doublecartoon_tov, equation, mesh)
    else
      TODO()
    end
  else
    if P.prms.problem === :bondi
      TODO()
      broadcast_bdry!(bdryllf_doublecartoon_bondi_mirrored, equation, mesh)
    elseif P.prms.problem === :tov
      impose_symmetry_bdry_values_doublecartoon!(mesh)
      broadcast_bdry!(bdryllf_doublecartoon_tov_mirrored, equation, mesh)
    else
      TODO()
    end
  end

  broadcast_volume!(av_flux_covariant_doublecartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, ldg_rD,       bdry_ldg_rD)
  dg1d.interpolate_face_data!(mesh, ldg_rSx,      bdry_ldg_rSx)
  dg1d.interpolate_face_data!(mesh, ldg_rτ,       bdry_ldg_rτ)
  dg1d.interpolate_face_data!(mesh, rhs_rD,       bdry_rhs_rD)
  dg1d.interpolate_face_data!(mesh, rhs_rSx,      bdry_rhs_rSx)
  dg1d.interpolate_face_data!(mesh, rhs_rτ,       bdry_rhs_rτ)
  dg1d.interpolate_face_data!(mesh, smoothed_mu,  bdry_smoothed_mu)

  broadcast_faces!(av_nflux_covariant_doublecartoon, equation, mesh)

  if :rD ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rD,  flx_rDx,  src_rD,  nflx_rD,  mesh)
  end
  if :rSx ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rSx, flx_rSxx, src_rSx, nflx_rSx, mesh)
  end
  if :rτ ∉ P.prms.freeze_vars
    compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  src_rτ,  nflx_rτ,  mesh)
  end

  if !P.prms.atm_evolve
    broadcast_volume!(stop_atmosphere_evolution_doublecartoon, P.equation, mesh)
  end

  return
end


#######################################################################
#                               Mesh2d                                #
#######################################################################


function rhs!(mesh::Mesh2d, P::Project{:cartoon})

  @unpack cache = mesh
  @unpack equation = P

  @unpack rD, rSx, rSz, rτ                      = get_dynamic_variables(cache)
  @unpack rhs_rD, rhs_rSx, rhs_rSz, rhs_rτ      = get_rhs_variables(cache)
  @unpack flx_rDx, flx_rDz,
          flx_rSxx, flx_rSxz,
          flx_rSzx, flx_rSzz,
          flx_rτx, flx_rτz,
          D, Sx, Sz, τ, max_v, vx, vz, p,
          src_rD, src_rSx, src_rSz, src_rτ      = get_static_variables(cache)
  @unpack nflx_rD, nflx_rSx, nflx_rSz, nflx_rτ,
          bdry_D, bdry_Sx, bdry_Sz, bdry_τ,
          bdry_max_v,  bdry_vx, bdry_vz, bdry_p = get_bdry_variables(cache)

  broadcast_volume!(unscale_conservatives_cartoon, equation, mesh)
  if P.prms.c2p_enforce_causal_atm || P.prms.c2p_enforce_atm
    broadcast_volume!(cons2prim_cartoon_freeze_flags, equation, mesh)
  else
    broadcast_volume!(cons2prim_cartoon, equation, mesh)
  end
  broadcast_volume!(maxspeed_cartoon, equation, mesh)

  dg1d.interpolate_face_data!(mesh, D,     bdry_D)
  dg1d.interpolate_face_data!(mesh, Sx,    bdry_Sx)
  dg1d.interpolate_face_data!(mesh, Sz,    bdry_Sz)
  dg1d.interpolate_face_data!(mesh, τ,     bdry_τ)
  dg1d.interpolate_face_data!(mesh, max_v, bdry_max_v)
  dg1d.interpolate_face_data!(mesh, vx,    bdry_vx)
  dg1d.interpolate_face_data!(mesh, vz,    bdry_vz)
  dg1d.interpolate_face_data!(mesh, p,     bdry_p)

  impose_symmetry_cartoon!(P, mesh)
  broadcast_volume!(flux_source_cartoon, equation, mesh)
  broadcast_faces!(llf_cartoon, equation, mesh)
  # assuming reflection symmetry for now
  impose_symmetry_bdry_values_cartoon!(mesh)
  broadcast_bdry!(bdryllf_cartoon_tov_reflected, equation, mesh)

  compute_rhs_weak_form!(rhs_rD,  flx_rDx,  flx_rDz,  src_rD,  nflx_rD,  mesh)
  compute_rhs_weak_form!(rhs_rSx, flx_rSxx, flx_rSxz, src_rSx, nflx_rSx, mesh)
  compute_rhs_weak_form!(rhs_rSz, flx_rSzx, flx_rSzz, src_rSz, nflx_rSz, mesh)
  compute_rhs_weak_form!(rhs_rτ,  flx_rτx,  flx_rτz,  src_rτ,  nflx_rτ,  mesh)

  if !P.prms.atm_evolve
    broadcast_volume!(determine_atmosphere, P.equation, mesh)
    broadcast_volume!(stop_atmosphere_evolution_cartoon, P.equation, mesh)
  end

  return
end
