module GLGL


using LinearAlgebra, Jacobi, ToggleableAsserts
import dg1d: purge_zeros!, lagrange_poly


"""
Computes quadrature nodes, weights and derivative weights
for the Generalized-Gauss-Lobatto quadrature rule.
n_pts ... number of quadrature nodes
Return values
x ... quadrature nodes, vector
w ... quadrature weights, vector
v ... derivative weight, number
"""
function rule(n_pts::Integer)

  N = n_pts - 2
  gb_x, gb_w = [], []
  if N > 0
    # Integration nodes and weights for Gauss-Gegenbauer quadrature
    α = 2; β = 2
    gb_x = zgj(N,    α, β)
    gb_w = wgj(gb_x, α, β)
  end

  x = zeros(n_pts)
  w = zeros(n_pts)

  x[2:end-1] .= gb_x
  x[1] = -1.0
  x[end] = 1.0

  w[2:end-1] .= @. gb_w / (1.0 - gb_x^2)^2

  w[1] = w[end] = 8.0 / 3.0 * (2.0 * N^2 + 10.0 * N + 9.0) / ((N+1.0) * (N+2.0) * (N+3.0) * (N+4.0))
  v = 8.0 / ((N+1.0) * (N+2.0) * (N+3.0) * (N+4.0))

  return x, w, v
end # function



"""
  integrate(f, x, w, v, D)

Integrate `f` sampled on GLGL quadrature nodes `x` with quadrature weights `w`,
boundary weight `v` and derivative matrix `D` based on `x`.
Use `rule` to obtain `x, w, v` and use `first_derivative_matrix` to obtain `D`.
"""
function integrate(f::AbstractVector, x::AbstractVector, w::AbstractVector, v::Real, D::AbstractVector)

  N = length(w)
  @toggled_assert (N,N) == size(D)
  @toggled_assert N == length(f)

  sum = dot(f, w)
  sum += v * dot(view(D,1,:), f)
  sum -= v * dot(view(D,N,:), f)

  return sum
end


"""
  integrate(f1, f2, w, v, D)

Similar to `integrate(f, w)`, but computes the inner product of `f1*f2` where both
`f1, f2` must be sampled on the GLGL quadrature nodes `x` associated with the quadrature
weights `w`, boundary weight `v` and derivative matrix `D` based on `x`.
Use `rule` to obtain `x, w, v` and use `first_derivative_matrix` to obtain `D`.
"""
function integrate(f1::AbstractVector, f2::AbstractVector, x::AbstractVector, w::AbstractVector,
    v::Real, D::AbstractMatrix)

  N = length(w)
  @toggled_assert (N,N) == size(D)
  @toggled_assert N == length(f1)
  @toggled_assert N == length(f2)

  sum = dot(f1.*f2, w)
  vD1, vDN = view(D,1,:), view(D,N,:)
  sum += v * (f1[1] * dot(vD1, f2) + f2[1] * dot(vD1, f1))
  sum -= v * (f1[end] * dot(vDN, f2) + f2[end] * dot(vDN, f1))

  return sum
end


"""
  integrate(f::Function, idxs::AbstractVector{<:Integer}, x, w)

Integrate `f` sampled on GLGL quadrature nodes `x` with quadrature weights `w`,
boundary weight `v` and derivative matrix `D` based on `x`.
`f(idxs[i])` shall return the value of the integrand evaluated at `x[i]`.
Use `rule` to obtain `x, w, v` and use `first_derivative_matrix` to obtain `D`.
"""
function integrate(fn::Function, idxs::AbstractVector{<:Integer}, x::AbstractVector, w::AbstractVector,
    v::Real, D::AbstractMatrix)

  N = length(w)
  @toggled_assert (N,N) == size(D)
  @toggled_assert N == length(idxs)

  sum = 0.0
  for i in 1:N
    fi = fn(idxs[i])
    sum += (w[i] + v * (D[1,i] - D[N,i])) * fi
  end

  return sum
end


"""
  integrate(f::Function, x, w, v, D)

Integrate `f` sampled on GLGL quadrature nodes `x` with quadrature weights `w`,
boundary weight `v` and derivative matrix `D` based on `x`.
`f(x[i])` shall return the value of the integrand evaluated at `x[i]`.
Use `rule` to obtain `x, w, v` and use `first_derivative_matrix` to obtain `D`.

This is similar to `integrate(f, idxs, x, w, v, D)`, but is better suited if the integrand is
given in closed form.
"""
function integrate(f::Function, x::AbstractVector, w::AbstractVector, v::Real, D::AbstractMatrix)
  @toggled_assert length(w) == length(x) == size(D, 1) == size(D, 2)
  sum = 0.0
  N = length(w)
  for i in 1:N
    fi = f(x[i])
    sum += (w[i] + v * (D[1,i] - D[N,i])) * fi
  end
  return sum
end


"""
M_ij = (li, lj)
x ... GLGL nodes
w ... GLGL weights
v ... GLGL derivative weight
D ... Derivative matrix based on x.
"""
function mass_matrix(x::AbstractVector, w::AbstractVector, v::Real, D::AbstractMatrix)

  N = length(x)
  @assert N == length(w)
  @assert (N,N) == size(D)

  M = diagm(w)
  M[1,:] += v * D[1,:]
  M[:,1] += v * D[1,:]
  M[end,:] -= v * D[end,:]
  M[:,end] -= v * D[end,:]

  return M
end # function



"""
S_ij = (li, dlj/dx)
Exactly representable, because product is of order 2N - 1 which can be integrated
exactly using LGL quadrature.
x ... GLGL nodes
w ... GLGL weights
v ... GLGL derivative weight
D ... First derivative matrix based on x.
D2 ... Second derivative matrix based on x.
"""
function stiffness_matrix(x::AbstractVector, w::AbstractVector, v::Real,
    D::AbstractMatrix, D2::AbstractMatrix)

  N = length(x)
  @assert N == length(w)
  @assert (N,N) == size(D)
  @assert (N,N) == size(D2)

  S = w .* D
  S[1,:]    += v * D2[1,:]
  S[end,:]  -= v * D2[end,:]
  S .+= v * D[1,:] .* transpose(D[1,:])
  S .-= v * D[end,:] .* transpose(D[end,:])

  return S
end


# TODO Add versions of mass_matrix, stiffness_matrix that take
# a single argument n_pts.


"""
L_ij = (P_i, P_j)
Analogon to mass matrix of Lagrange polynomials, expresses the linear dependency
of the numerical representation of the Legendre basis. Ideally, it is a
diagonal matrix (which should be ensured by the GLGL quadrature).
x ... GLGL nodes
w ... GLGL weights
v ... GLGL derivative weight
D ... First derivative matrix based on x.
"""
function linear_dependency_matrix(x::AbstractVector, w::AbstractVector, v::Real, D::AbstractMatrix)
  N = length(x)
  L = zeros(N,N)
  for i in 1:N
    Pi = map(x->legendre(x, i-1), x)
    for j in i:N
      Pj = map(x->legendre(x, j-1), x)
      lij = integrate(Pi, Pj, x, w, v, D)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  purge_zeros!(L)

  return L
end


"""
E_ij = (l_i, P_j)
Commonly used to project either Lagrange polynomials onto Legendres or
reverse.
x ... GLGL nodes
w ... GLGL weights
v ... GLGL derivative weight
D ... First derivative matrix based on x.
"""
function expansion_matrix(x::AbstractVector, w::AbstractVector, v::Real, D::AbstractMatrix)
  N = length(x)
  T = zeros(N,N)
  for i in 1:N
    li = map(xi->lagrange_poly(i, xi, x), x)
    for j in 1:N
      Pj = map(x->legendre(x, j-1), x)
      tij = integrate(li, Pj, x, w, v, D)
      T[i,j] = tij
    end
  end
  purge_zeros!(T)

  return T
end



end # module
