Base.@kwdef struct SpectralElement
  N::Int64                # polynomial order
  Npts::Int64             # number of quadrature points = N + 1
  z::Vector{Float64}      # quadrature points
  w::Vector{Float64}      # quadrature weights
  v::Float64              # derivative weight (only needed for GLGL quadrature rule)
  bw::Vector{Float64}     # barycentric weights
  D::Matrix{Float64}      # derivative matrix
  S::Matrix{Float64}      # stiffness matrix
  invM::Matrix{Float64}   # inverse mass matrix
  MDM::Matrix{Float64}    # = (M^-1 D^T M) = M^-1 S^T
  Ml_lhs::Vector{Float64} # = M^-1 l(-1)
  Ml_rhs::Vector{Float64} # = M^-1 l(+1)
  V::Matrix{Float64}      # = P_(j-1)(z_i), Vandermonde matrix
  invV::Matrix{Float64}   # inv(V), inverse Vandermonde matrix
  quadr::Symbol           # :LGL, :LGL_mass_lumped, :GLGL
  F::Matrix{Float64}      # spectral filtering matrix

  function SpectralElement(N::Int, quadr::Symbol;
      # filter parameters
      ηc::Float64=0.5, sh::Float64=16.0, α::Float64=36.0)
    Npts = N + 1

    if quadr === :LGL
      z, w        = LGL.rule(Npts)
      v           = -1.0 # unused here
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_exact(Npts)
      # TODO Figure out how much the aliasing error influences the accuracy of the inner product.
      # Could be totally dealised using the Orszag 2/3rds rule.
      L           = LGL.linear_dependency_matrix_exact(z, w)
      E           = LGL.expansion_matrix_exact(z, w)
    elseif quadr === :LGL_lumped
      z, w        = LGL.rule(Npts)
      v           = -1.0 # unused here
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_lumped(Npts)
      L           = LGL.linear_dependency_matrix_lumped(z, w)
      E           = LGL.expansion_matrix_lumped(z, w)
    elseif quadr === :GLGL
      z, w, v     = GLGL.rule(Npts)
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      D2          = second_derivative_matrix(z)
      M           = GLGL.mass_matrix(z, w, v, D)
      S           = GLGL.stiffness_matrix(z, w, v, D, D2)
      invM        = inv(M)
      L           = GLGL.linear_dependency_matrix(z, w, v, D)
      E           = GLGL.expansion_matrix(z, w, v, D)
    else
      error("Invalid quadrature rule specified! Use one of :LGL_lumped, :LGL, :GLGL.")
    end

    MDM     = invM * transpose(S)
    purge_zeros!(MDM)
    V       = vandermonde_matrix(z) # TODO needs to be replaced with vandermonde_matrix(z, basis_fn)
    invV    = inv(V)
    Λ = diagm([ spectral_filter((i-1)/N, (ηc,sh,α)) for i = 1:Npts ])
    F = V * Λ * invV

    l_lhs   = zeros(N+1)
    l_rhs   = zeros(N+1)
    l_lhs[1]    = 1.0
    l_rhs[end]  = 1.0
    Ml_lhs = invM * l_lhs
    Ml_rhs = invM * l_rhs

    return new(N, Npts, z, w, v, bw, D, S, invM, MDM, Ml_lhs, Ml_rhs,
               V, invV, quadr, F)
  end
end


# (5.16) from Hesthaven, Warburton 2007
function spectral_filter(η, prms)
  ηc, sh, α = prms
  return if 0 ≤ η ≤ ηc
    1
  else
    exp(-α*((η-ηc)/(1-ηc))^(2*sh))
  end
end


@inline function inner_product(u1, u2, el::SpectralElement)
  @unpack quadr = el
  @toggled_assert quadr in (:LGL, :LGL_lumped, :GLGL)
  if quadr === :LGL || quadr === :LGL_lumped
    return LGL.integrate(u1, u2, el.z, el.w)
  else # quadr === :GLGL
    return GLGL.integrate(u1, u2, el.z, el.w, el.v, el.D)
  end
end


@inline function integrate(u, el::SpectralElement)
  @unpack quadr = el
  @toggled_assert quadr in (:LGL, :LGL_lumped, :GLGL)
  if quadr === :LGL || quadr === :LGL_lumped
    return LGL.integrate(u, el.z, el.w)
  else # quadr === :GLGL
    return GLGL.integrate(u, el.z, el.w, el.v, el.D)
  end
end


@inline function integrate(fn::Function, idxs::AbstractVector{<:Integer}, el::SpectralElement)
  @unpack quadr = el
  @toggled_assert quadr in (:LGL, :LGL_lumped, :GLGL)
  if quadr === :LGL || quadr === :LGL_lumped
    return LGL.integrate(fn, idxs, el.z, el.w)
  else # quadr === :GLGL
    return GLGL.integrate(fn, idxs, el.z, el.w, el.v, el.D)
  end
end


@inline function differentiate(u, el::SpectralElement)
  return el.D * u
end

@inline function differentiate!(du, u, el::SpectralElement)
  mul!(du, el.D, u)
end
